@@ -20,6 +20,8 @@ import os
 # TODO - possible integration with mirdata
 # TODO - optionally avoud splitting notes...
 #      - see magenta/models/onsets_frames_transcription/audio_label_data_utils.py
+# TODO - multipitch can overlap on last/first frame of notes
+#        - seems correct based on note start/end times but could be problematic
 
 # TODO - get notes function?
 
@@ -30,7 +32,7 @@ class TranscriptionDataset(Dataset):
     """
 
     def __init__(self, base_dir, splits, hop_length, sample_rate, data_proc, profile, num_frames,
-                 split_notes, reset_data, store_data, save_data, save_loc, seed):
+                 audio_norm, split_notes, reset_data, store_data, save_data, save_loc, seed):
         """
         Initialize parameters common to all datasets as fields and instantiate
         as a PyTorch Dataset.
@@ -51,6 +53,11 @@ class TranscriptionDataset(Dataset):
           Instructions for organizing data and ground-truth
         num_frames : int
           Number of frames per data sample
+        audio_norm : float or None
+          Type of normalization to perform when loading audio
+          -1 - root-mean-square
+          See librosa for others...
+            - None case is handled here
         split_notes : bool
           Flag to avoiding cutting samples in between notes
         reset_data : bool
@@ -87,6 +94,8 @@ class TranscriptionDataset(Dataset):
 
         # Default the feature extraction to a plain Mel Spectrogram if none was provided
         if data_proc is None:
+            # TODO - should keep as None so that, e.g., pure tablature data won't do this step
+            # TODO - or create placeholders such as Waveform() or Frames()
             data_proc = STFT(hop_length=self.hop_length,
                              sample_rate=self.sample_rate)
         self.data_proc = data_proc
@@ -106,6 +115,8 @@ class TranscriptionDataset(Dataset):
         # Set the number of frames for each sample
         self.num_frames = num_frames
 
+        self.audio_norm = audio_norm
+
         # Set the storing and saving parameters
         self.store_data = store_data
         self.save_data = save_data
@@ -117,14 +128,16 @@ class TranscriptionDataset(Dataset):
         # Remove any saved ground-truth for the dataset if reset_data is selected
         if os.path.exists(self.get_gt_dir()) and self.reset_data:
             shutil.rmtree(self.get_gt_dir())
-        # Make sure the directory for saving and loading ground-truth exists
-        os.makedirs(self.get_gt_dir(), exist_ok=True)
+        if self.save_data:
+            # Make sure the directory for saving and loading ground-truth exists
+            os.makedirs(self.get_gt_dir(), exist_ok=True)
 
         # Remove any saved features for the dataset if reset_data is selected
         if os.path.exists(self.get_feats_dir()) and self.reset_data:
             shutil.rmtree(self.get_feats_dir())
-        # Make sure the directory for saving and loading features exists
-        os.makedirs(self.get_feats_dir(), exist_ok=True)
+        if self.save_data:
+            # Make sure the directory for saving and loading features exists
+            os.makedirs(self.get_feats_dir(), exist_ok=True)
 
         # Initialize a random number generator for the dataset
         self.rng = np.random.RandomState(seed)
@@ -176,7 +189,7 @@ class TranscriptionDataset(Dataset):
         data = self.get_track_data(track_id)
 
         # Convert all numpy arrays in the data dictionary to float32
-        data = tools.track_to_dtype(data, dtype=tools.FLOAT32)
+        data = tools.dict_to_dtype(data, dtype=tools.FLOAT32)
 
         # Remove any notes, as they cannot be batched
         if tools.query_dict(data, tools.KEY_NOTES):
@@ -247,20 +260,27 @@ class TranscriptionDataset(Dataset):
         if self.sample_rate != fs or self.hop_length != hop_length:
             warnings.warn('Loaded features\' sampling rate or hop length differs from expected.')
 
-        # Calculate the frame times every time (faster than saving/loading)
-        times = self.data_proc.get_times(data[tools.KEY_AUDIO])
+        if tools.query_dict(data, tools.KEY_TIMES):
+            # Use the times that were already provided
+            times = data[tools.KEY_TIMES]
+        else:
+            # Calculate the frame times (faster than saving/loading)
+            times = self.data_proc.get_times(data[tools.KEY_AUDIO])
+            # Add the times to the data dictionary
+            data[tools.KEY_TIMES] = times
 
         # Check if fixed features were provided
         if feats is not None:
             # Add the features to the data dictionary
             data[tools.KEY_FEATS] = feats
-        data[tools.KEY_TIMES] = times
 
         if self.store_data:
             # Check if fixed features were provided
             if feats is not None:
                 # Add the features to the data dictionary in RAM
                 self.data[track][tools.KEY_FEATS] = feats
+                # TODO - a lot of memory could be saved by throwing
+                #        away audio after features are computed
             self.data[track][tools.KEY_TIMES] = times
 
         return data
@@ -331,7 +351,7 @@ class TranscriptionDataset(Dataset):
         if tools.KEY_NOTES in keys:
             # Notes entry has not been popped by a dataloader
             notes = data[tools.KEY_NOTES]
-        elif tools.KEY_NOTES in self.data[track_id].keys():
+        elif self.store_data and tools.KEY_NOTES in self.data[track_id].keys():
             # Notes entry has been popped and must be added again
             notes = self.data[track_id][tools.KEY_NOTES]
         else:
