@@ -21,7 +21,7 @@ KEY_AUDIO
 """
 
 from .constants import *
-from .instrument import *
+from .instrument import InstrumentProfile, PianoProfile, TablatureProfile, GuitarProfile
 from .io import *
 from .utils import *
-#from .visualize import *
+from .visualize import *
