@@ -16,6 +16,29 @@ import mido
 import jams
 import os
 
+__all__ = [
+    'load_normalize_audio',
+    'extract_stacked_notes_jams',
+    'load_stacked_notes_jams',
+    'extract_notes_jams',
+    'load_notes_jams',
+    'extract_duration_jams',
+    'load_duration_jams',
+    'load_stacked_pitch_list_jams',
+    'load_pitch_list_jams',
+    'load_notes_midi',
+    'write_and_print',
+    'write_list',
+    'write_pitch_list',
+    'write_notes',
+    'write_stacked_notes_jams',
+    'stream_url_resource',
+    'unzip_and_remove',
+    'zip_and_save',
+    'change_base_dir',
+    'file_sort'
+]
+
 
 ##################################################
 # INPUT                                          #
@@ -32,10 +55,11 @@ def load_normalize_audio(wav_path, fs=None, norm=-1, res_type='kaiser_best'):
       Path to audio file to read
     fs : int or float or None (optional)
       Desired sampling rate
-    norm : float
+    norm : float or None
       Type of normalization to perform
       -1 - root-mean-square
       See librosa for others...
+        - None case is handled here
     res_type : string
       See librosa... - this significantly affects the speed of resampling long audio files
 
@@ -56,14 +80,59 @@ def load_normalize_audio(wav_path, fs=None, norm=-1, res_type='kaiser_best'):
         audio = utils.rms_norm(audio)
     else:
         # Normalize the audio using librosa
-        audio = librosa.util.normalize(audio, norm)
+        audio = librosa.util.normalize(audio, norm=norm)
 
     return audio, fs
 
 
+def extract_stacked_notes_jams(jam):
+    """
+    Extract MIDI notes spread across slices (e.g. guitar strings) from JAMS data into a dictionary.
+
+    Parameters
+    ----------
+    jam : JAMS object
+      JAMS file data
+
+    Returns
+    ----------
+    stacked_notes : dict
+      Dictionary containing (slice -> (pitches, intervals)) pairs
+    """
+
+    # Extract all of the midi note annotations
+    note_data_slices = jam.annotations[constants.JAMS_NOTE_MIDI]
+
+    # Initialize a dictionary to hold the notes
+    stacked_notes = dict()
+
+    # Loop through the slices of the stack
+    for slice_notes in note_data_slices:
+
+        # Extract the string label for this slice
+        string = slice_notes.annotation_metadata[constants.JAMS_STRING_IDX]
+
+        # Initialize lists to hold the pitches and intervals
+        pitches, intervals = list(), list()
+
+        # Loop through the notes pertaining to this slice
+        for note in slice_notes:
+            # Append the note pitch and interval to the respective list
+            pitches.append(note.value)
+            intervals.append([note.time, note.time + note.duration])
+
+        # Convert the pitch and interval lists to arrays
+        pitches, intervals = np.array(pitches), np.array(intervals)
+
+        # Add the pitch-interval pairs to the stacked notes dictionary under the string entry as key
+        stacked_notes.update(utils.notes_to_stacked_notes(pitches, intervals, string))
+
+    return stacked_notes
+
+
 def load_stacked_notes_jams(jams_path):
     """
-    Load MIDI notes spread across slices (e.g. guitar strings) from JAMS file into a dictionary.
+    Helper function to load a JAMS file and extract the stacked notes.
 
     Parameters
     ----------
@@ -79,41 +148,43 @@ def load_stacked_notes_jams(jams_path):
     # Load the metadata from the jams file
     jam = jams.load(jams_path)
 
-    # Extract all of the midi note annotations
-    note_data_slices = jam.annotations[constants.JAMS_NOTE_MIDI]
-
-    # Obtain the number of annotations
-    stack_size = len(note_data_slices)
-
-    # Initialize a dictionary to hold the notes
-    stacked_notes = dict()
-
-    # Loop through the slices of the stack
-    for slc in range(stack_size):
-        # Extract the notes pertaining to this slice
-        slice_notes = note_data_slices[slc]
-
-        # Initialize lists to hold the pitches and intervals
-        pitches, intervals = list(), list()
-
-        # Loop through the notes pertaining to this slice
-        for note in slice_notes:
-            # Append the note pitch and interval to the respective list
-            pitches.append(note.value)
-            intervals.append([note.time, note.time + note.duration])
-
-        # Convert the pitch and interval lists to arrays
-        pitches, intervals = np.array(pitches), np.array(intervals)
-
-        # Add the pitch-interval pairs to the stacked notes dictionary under the slice key
-        stacked_notes.update(utils.notes_to_stacked_notes(pitches, intervals, slc))
+    # Extract the stacked notes
+    stacked_notes = extract_stacked_notes_jams(jam)
 
     return stacked_notes
 
 
+def extract_notes_jams(jam):
+    """
+    Extract all MIDI notes within a JAMS file.
+
+    Parameters
+    ----------
+    jam : JAMS object
+      JAMS file data
+
+    Returns
+    ----------
+    pitches : ndarray (N)
+      Array of pitches corresponding to notes
+      N - number of notes
+    intervals : ndarray (N x 2)
+      Array of onset-offset time pairs corresponding to notes
+      N - number of notes
+    """
+
+    # First, extract the notes into a stack
+    stacked_notes = extract_stacked_notes_jams(jam)
+
+    # Unpack the stacked notes
+    pitches, intervals = utils.stacked_notes_to_notes(stacked_notes)
+
+    return pitches, intervals
+
+
 def load_notes_jams(jams_path):
     """
-    Load all MIDI notes within a JAMS file.
+    Helper function to load a JAMS file and extract the notes.
 
     Parameters
     ----------
@@ -130,17 +201,61 @@ def load_notes_jams(jams_path):
       N - number of notes
     """
 
-    # First, load the notes into a stack
-    stacked_notes = load_stacked_notes_jams(jams_path=jams_path)
+    # Load the metadata from the jams file
+    jam = jams.load(jams_path)
 
-    pitches, intervals = utils.stacked_notes_to_notes(stacked_notes)
+    # Extract the notes as loose groups
+    pitches, intervals = extract_notes_jams(jam)
 
     return pitches, intervals
 
 
+def extract_duration_jams(jam):
+    """
+    Extract the duration of audio associated with JAMS annotations.
+
+    Parameters
+    ----------
+    jam : JAMS object
+      JAMS file data
+
+    Returns
+    ----------
+    duration : float
+      Total length (seconds) of the audio associated with the annotations
+    """
+
+    # Read the meta-data from the jams file
+    duration = jam[constants.JAMS_METADATA].duration
+
+    return duration
+
+
+def load_duration_jams(jams_path):
+    """
+    Helper function to load a JAMS file and extract the duration.
+
+    Parameters
+    ----------
+    jams_path : string
+      Path to JAMS file to read
+
+    Returns
+    ----------
+    duration : float
+      Total length (seconds) of the audio associated with the annotations
+    """
+
+    # Read the meta-data from the jams file
+    duration = extract_duration_jams(jams.load(jams_path))
+
+    return duration
+
+
 def load_stacked_pitch_list_jams(jams_path, times=None):
     """
     Load pitch lists spread across slices (e.g. guitar strings) from JAMS file into a dictionary.
+    TODO - same load/extract separation as notes
 
     Parameters
     ----------
@@ -213,6 +328,7 @@ def load_stacked_pitch_list_jams(jams_path, times=None):
 def load_pitch_list_jams(jams_path, times):
     """
     Load pitch list from JAMS file.
+    TODO - same load/extract separation as notes
 
     Parameters
     ----------
@@ -368,6 +484,9 @@ def write_and_print(file, text, verbose=True, end=''):
       Append this to the end of the text (e.g. for new line or no new line)
     """
 
+    # Make sure the provided text is a string
+    text = str(text)
+
     # Append the ending to the text
     text = text + end
 
@@ -381,6 +500,31 @@ def write_and_print(file, text, verbose=True, end=''):
             print(text, end='')
 
 
+def write_list(lst, path):
+    """
+    Write all items in a list to a file.
+
+    Parameters
+    ----------
+    lst : list of object
+      Collection of items to write
+    path : string
+      Location of the file to write
+    """
+
+    # Make sure all the directories in the path exist
+    os.makedirs(os.path.dirname(path), exist_ok=True)
+
+    # Open a file at the path with writing permissions
+    with open(path, 'w') as file:
+        # Loop through the entire list
+        for i, item in enumerate(lst):
+            # Determine how the line should end
+            end = '' if (i + 1) == len(lst) else '\n'
+            # Write the line to the file
+            write_and_print(file, item, verbose=False, end=end)
+
+
 def write_pitch_list(times, pitches, path, places=3):
     """
     Write all of the notes in a loose collection to a file.
@@ -458,6 +602,67 @@ def write_notes(pitches, intervals, path, places=3):
             write_and_print(estim_file, line, verbose=False, end=end)
 
 
+def write_stacked_notes_jams(stacked_notes, jams_path):
+    """
+    Helper function to create a JAMS file and populate it with stacked notes.
+
+    Parameters
+    ----------
+    stacked_notes : dict
+      Dictionary containing (slice -> (pitches, intervals)) pairs
+    jams_path : string
+      Path to JAMS file to write
+    """
+
+    # Create a new JAMS object
+    jam = jams.JAMS()
+
+    # Keep track of the duration
+    total_duration = 0
+
+    # Loop through all slices
+    for key in stacked_notes.keys():
+        # Initialize a new annotation to hold the notes of the slice
+        slice_data = jams.Annotation(namespace=constants.JAMS_NOTE_MIDI, time=0, duration=0)
+        # Add metadata to reference the slice as the source of data
+        slice_data.annotation_metadata = jams.AnnotationMetadata(data_source=key)
+
+        # Extract the notes corresponding to the slice
+        pitches, intervals = stacked_notes[key]
+
+        # Compute the duration of each note
+        durations = intervals[:, 1] - intervals[:, 0]
+
+        # Loop through all the notes
+        for n in range(len(pitches)):
+            # Add the note to the slice data
+            slice_data.append(time=intervals[n, 0], duration=durations[n], value=pitches[n])
+
+        # Add the annotation to the JAM
+        jam.annotations.append(slice_data)
+
+        if len(pitches) == 0:
+            # Don't compute duration if there are no notes for the slice
+            continue
+
+        # Determine the total duration of the slice
+        slice_duration = np.max(intervals[:, 1])
+
+        # Update the slice duration, if necessary
+        if slice_duration > total_duration:
+            total_duration = slice_duration
+
+    # Add the total duration to the file metadata
+    jam.file_metadata.duration = total_duration
+
+    # Add the total duration to the metadata of each annotation
+    for slice_data in jam.annotations:
+        slice_data.duration = total_duration
+
+    # Save as a JAM file
+    jam.save(jams_path)
+
+
 ##################################################
 # USEFUL FILE MANAGEMENT                         #
 ##################################################
@@ -522,6 +727,31 @@ def unzip_and_remove(zip_path, target=None):
     os.remove(zip_path)
 
 
+def zip_and_save(dir_path, zip_path):
+    """
+    Zip the contents of a directory and save the compressed file to disk.
+
+    Parameters
+    ----------
+    dir_path : string
+      Path to the directory to compress
+    zip_path : string
+      Path to the resulting zip file
+    """
+
+    # Create a new Zip file to write to
+    with zipfile.ZipFile(zip_path, mode='w') as zipf:
+        # Traverse through the directory to compress
+        for root, _, files in os.walk(dir_path):
+            # Loop through all files in the current sub-directory
+            for file in files:
+                # Determine the absolute and relative path of the file
+                absolute_path = os.path.join(root, file)
+                relative_path = absolute_path.replace(dir_path, '')
+                # Write to the Zip file
+                zipf.write(absolute_path, relative_path)
+
+
 def change_base_dir(new_dir, old_dir):
     """
     Change the top-level directory from the path chain of each file.
@@ -545,3 +775,26 @@ def change_base_dir(new_dir, old_dir):
 
     # Remove the (now empty) old top-level directory
     os.rmdir(old_dir)
+
+
+def file_sort(file_name):
+    """
+    Augment file names for sorting since, e.g., /'500/' will by default be
+    scored as higher than /'1500/'. One way to fix this is by adding the
+    length of the file to the beginning of the string.
+
+    Parameters
+    ----------
+    file_name : str
+      Path being sorted
+
+    Returns
+    ----------
+    sort_name : str
+      Character count concatenated with original file name
+    """
+
+    # Takes into account the number of digits by adding string length
+    sort_name = str(len(file_name)) + file_name
+
+    return sort_name
